unit md2;

interface
 uses
  SysUtils, Classes,
  {$ifdef fpc}
    un_typedefs in '../../cge/src/un_typedefs.pas',
    mo_hub in '../../cge/src/mo_hub.pas',
    mo_classes in '../../cge/src/mo_classes.pas',
    mo_textures in '../../cge/src/mo_textures.pas',
    mo_module in '../../cge/src/mo_module.pas',
    mo_controls in '../../cge/src/mo_controls.pas'
  {$else}
    un_typedefs,
    mo_hub,
    mo_classes,
    mo_textures,
    mo_module,
    mo_controls
  {$endif}
  ;

const
  MAX_TRIANGLES = 4096;
  MAX_VERTS = 2048;
  MAX_FRAMES = 512;
  MAX_MD2SKINS = 32;
  MAX_SKINNAME = 64;

type

  TMd2Header = packed record
    ident: Integer;
    version: Integer;

    skinwidth: Integer;
    skinheight: Integer;
    framesize: Integer;

    NumSkins: Integer;
    NumVertices: Integer;
    num_st: Integer;
    num_tris: Integer;
    NumGLCommands: Integer;
    NumFrames: Integer;

    OffsetSkins: Integer;
    ofs_st: Integer;
    ofs_tris: Integer;
    OffsetFrames: Integer;
    OffsetGLCommands: Integer;
    ofs_end: Integer;
  end;

  TMyVertex = packed record
    x, y, z: float;
    lightnormalindex: integer;
  end;

  TGLVertex = packed record
    s, t: float;
    Vertind: cardinal;
  end;

  TGlCommand = packed record
    Command: (gcFan, gcStrip);
    Vert: array of TGLVertex;
  end;

  TAnimRec = record
    AID, SEL: integer;
    SEQfnum: array of integer;
    SEQPos: array of float;
  end;

Type
  TMD2 = class
    Constructor Create(FileName: string);
    Destructor Destroy; OVERRIDE;
    procedure _Draw(Frame: integer); OVERLOAD;
    procedure _Draw(Frame1, Frame2: integer; Factor: single); OVERLOAD;
    Procedure _Markup;
    function FrameNum(name: string): integer;
  private
    fname: string;
    Header: TMd2Header;
    GLCommand: array of TGLCommand;
    Vertex: array of array of TMyVertex;
    FrameName,
    SkinName: array of String;
    Function _NumFrames: integer;
    Function _NumVertices: integer;
    Function _Numskins: integer;
  public
    property NumFrames: integer read _numframes;
    property NumVertices: integer read _numvertices;
    property Numskins: integer read _numskins;
  end;

  TMD2S = class
    Constructor Create(FileName: string);
    Destructor Destroy; OVERRIDE;
  public

  end;

implementation

{ TMD2 }


Type
  TRawVertex = packed record
    x, y, z, lightNormalIndex: byte;
  end;

Constructor TMD2.Create(FileName: string);
var
  f: TFileStream;
  i, j, n: integer;
  s: string;
  st: array [0..5] of float;
  rv: array of TRawVertex;
begin
  inherited Create;
  if VerboseLog then addlog('Loading "%0" ... ', [FileName]);
  fname:=FileName;
  FillChar(Header, SizeOf(Header), 0);
  f:= TFileStream.Create(Filename, fmOpenRead + fmShareDenyWrite);
  f.Read(Header, SizeOf(Header));
  //reading skin names
  //  
  f.Position:=Header.OffsetSkins;
  SetLength(SkinName, Header.NumSkins);
  For i:=0 to High(SkinName) do begin
    SetLength(s, 64);
    f.Read(s[1], 64);
    SkinName[i]:=UpperCase(String(PChar(s)));
  end;
  //reading and transforming GL commands
  //   GL 
  //(..        )
  f.Position:=Header.OffsetGLCommands;
  n:=0;
  Repeat
    f.Read(j, 4);
    If j=0 then BREAK;
    inc(n);
    SetLength(GLCommand, n);
    with GLCommand[n-1] do begin
      if j>0 then Command:=gcStrip
             else Command:=gcFan;
      j:=abs(j);
      SetLength(Vert, j);
      For i:=0 to High(Vert) do begin
        f.Read(Vert[i], 12);
        //Vert[i].s:=1-vert[i].s;
        Vert[i].t:=1-vert[i].t;
      end;
    end
  Until IOResult<>0 {false};
  //reading and transforming vertexes and frames
  //      
  f.Position:=Header.OffsetFrames;
  SetLength(FrameName, Header.NumFrames);
  SetLength(Vertex, Header.NumFrames);
  SetLength(rv, Header.NumVertices);
  For i:=0 to High(Vertex) do begin
    f.Read(st, 24);
    SetLength(s, 16);
    f.Read(s[1], 16);
    FrameName[i]:=String(PChar(s));
    f.Read(rv[0], 4*Header.NumVertices);
    SetLength(Vertex[i], Header.NumVertices);
    For j:=0 to High(rv) do
      With Vertex[i, j] do begin
        LightNormalIndex:=rv[i].LightNormalIndex;
        x:=rv[j].x * st[0] + st[3];
        y:=rv[j].y * st[1] + st[4];
        z:=rv[j].z * st[2] + st[5];
      end;
  end;
  For i:=0 to High(FrameName) do FrameName[i]:=UpperCase(FrameName[i]);
  SetLength(rv, 0);
  f.free;
  _Markup;
  if VerboseLog then AddLogOk;
end;


destructor TMD2.Destroy;
var
  i, j : Integer;
begin
  // ,   ,
  //    0.
  SetLength(SkinName, 0);
  SetLength(FrameName,0);
  For i:=0 to High(GLCommand) do
    SetLength(GLCommand[i].Vert, 0);
  SetLength(GLCommand, 0);
  For j:=0 to High(Vertex) do
    SetLength(Vertex[j], 0);
  SetLength(Vertex, 0);
  inherited;
end;

function TMD2._NumFrames: integer;
begin
  Result:=Header.NumFrames;
end;

function TMD2._Numskins: integer;
begin
  Result:=Header.NumSkins;
end;

function TMD2._NumVertices: integer;
begin
  Result:=Header.NumVertices;
end;

procedure TMD2._Draw(Frame1, Frame2: integer; Factor: single);
var
  factor2: single;
  i, j, u: integer;
begin
//AddLog('md2: %0, %1, %2',[Frame1, Frame2, Factor]);
if factor>1 then factor:=1;
if factor<0 then factor:=0;
  factor2:=1-factor;
  //Rendering primitives
  For u:=0 to high(GLCommand) do With GLCommand[u] do begin
    Case Command of
      gcFan: glBegin(GL_TRIANGLE_FAN);
      gcStrip: glBegin(GL_TRIANGLE_STRIP);
    end;
  Try
    For i:=0 to high(vert) do With vert[i] do begin
    //glColor4f(v->color);
      glTexCoord2f(s, 1-t);
      glVertex3f(
        Vertex[Frame1, vertind].y* factor2 + Vertex[Frame2, vertind].y*factor,
        Vertex[Frame1, vertind].z* factor2 + Vertex[Frame2, vertind].z*factor,
        Vertex[Frame1, vertind].x* factor2 + Vertex[Frame2, vertind].x*factor);
    end;
  Except
    Die('ERROR_MD2_RENDERING_FAILURE',[fname,Frame1,Frame2,Factor]);
  End;
    glEnd;
  end;
end;


procedure TMD2._Draw(Frame: integer);
var
  i, j, u: integer;
begin
//AddLog('md2: %0',[Frame]);
  //Rendering primitives
  For u:=0 to high(GLCommand) do With GLCommand[u] do begin
    Case Command of
      gcFan: glBegin(GL_TRIANGLE_FAN);
      gcStrip: glBegin(GL_TRIANGLE_STRIP);
    end;
  Try
    For i:=0 to high(vert) do With vert[i] do begin
    //glColor4f(v->color);
      glTexCoord2f(s, 1-t);
      glVertex3f(Vertex[Frame, vertind].y,
                 Vertex[Frame, vertind].z,
                 Vertex[Frame, vertind].x);
    end;
  Except
    Die('ERROR_MD2_RENDERING_FAILURE',[fname,Frame,'(none)','(none)']);
  End;
    glEnd;
  end;
end;

procedure TMD2._Markup;
begin

end;

function TMD2.FrameNum(name: string): integer;
var i: integer;
begin
  name:=UpperCase(name);
  Result:=0;
  For i:=0 to High(FrameName) do begin
    if FrameName[i]=name then begin
      Result:=i;
      Break;
    end;
  end;
end;

{ TMD2S }

constructor TMD2S.Create(FileName: string);
begin
  inherited Create;

end;

destructor TMD2S.Destroy;
begin

  inherited;
end;

end.
