{
    This file is part of the Cheb's Game Engine,
    Copyright (c) 2004-2006 by Anton Rzheshevski (chebmaster@mail.ru),
      and contains source of the image postprocessing routines.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 **********************************************************************}
 
unit mo_postprocess;

interface

 uses
  SysUtils,
  {$ifdef fpc}
    un_typedefs in '../../cge/src/un_typedefs.pas',
    mo_hub in '../../cge/src/mo_hub.pas',
    mo_classes in '../../cge/src/mo_classes.pas',
    mo_textures in '../../cge/src/mo_textures.pas',
    mo_module in '../../cge/src/mo_module.pas',
    mo_controls in '../../cge/src/mo_controls.pas'
  {$else}
    un_typedefs,
    mo_hub,
    mo_classes,
    mo_textures,
    mo_module,
    mo_controls
  {$endif}
  ;
  
procedure HalvAndBlurImage (tex1, tex2: glUint;
                            tex1size,
                            imagewidth, imageheight: integer;
                            brightness: glFloat;
                            blurPower: integer);
                            
procedure PostprocessOutput(tex1, tex2: glUint;
                            tex1size, tex2size,
                            image1width, image1height,
                            image2width, image2height: integer;
                            brightness1, brightness2,
                            blur1, blur2: glFloat);
                            
Const
   BLUR_OFFSET = 1.0;

implementation


// --------------------------------------------------------------------------//

  procedure HalvAndBlurImage (tex1, tex2: glUint;
                              tex1size,
                              imagewidth, imageheight: integer;
                              brightness: glFloat;
                              blurPower: integer);
  var
    d, xo, yo, dx, dy: glFloat;
    i: integer;
  begin
    glColor4f(brightness/blurPower, brightness/blurPower, brightness/blurPower, 1);
    glViewPort(0,0,imagewidth, imageheight);
   // glClearColor(0,0,0,0);
   // glClear(GL_COLOR_BUFFER_BIT);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glOrtho(0, imagewidth, 0, imageheight, 0 ,1);
    glDisable(GL_DEPTH_TEST);
    glDepthMask(GL_FALSE);
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, tex1);
    glDisable(GL_BLEND);
    glDisable(GL_ALPHA_TEST);
    d:=(1/tex1size);
    xo:=imagewidth/tex1size;
    yo:=imageheight/tex1size;
    For i:=1 to BlurPower do begin
      if i > 1 then begin
        glEnable(GL_BLEND);
        glBlendFunc(GL_ONE, GL_ONE);
      end;
      case i of
        1: begin
            if BlurPower = 2 then begin
              dx:=-d * BLUR_OFFSET; dy:=-d * BLUR_OFFSET;
            end
            else begin
              dx:=0; dy:=0;
            end;
           end;
        2: begin dx:=d * BLUR_OFFSET; dy:=d * BLUR_OFFSET end;
        3: begin dx:=-d * BLUR_OFFSET; dy:=-d * BLUR_OFFSET end;
        4: begin dx:=d * BLUR_OFFSET; dy:=-d * BLUR_OFFSET end;
        5: begin dx:=-d * BLUR_OFFSET; dy:=d  * BLUR_OFFSET end;
      end;
      glBegin(GL_QUADS);
        glTexCoord2f(0 + d + dx, 0 - d +dy);
        glVertex2f(0, 0); //bottom left corner
        glTexCoord2f(0 + d + dx, yo - d +dy);
        glVertex2f(0, imageheight div 2);
        glTexCoord2f(xo + d+ dx, yo - d +dy);
        glVertex2f(imagewidth div 2, imageheight div 2);
        glTexCoord2f(xo + d+ dx, 0 - d +dy);
        glVertex2f(imagewidth div 2, 0);
      glEnd;
    end;

    glBindTexture(GL_TEXTURE_2D, tex2);
    glCopyTexSubImage2d(GL_TEXTURE_2D, 0, 0, 0, 0, 0,
                                          imagewidth div 2, imageheight div 2);
    glDepthMask(GL_TRUE);
  end;
  
// --------------------------------------------------------------------------//


  procedure PostprocessOutput(tex1, tex2: glUint;
                              tex1size, tex2size,
                              image1width, image1height,
                              image2width, image2height: integer;
                              brightness1, brightness2,
                              blur1, blur2: glFloat);
  var
    i: integer;
    d, xo, yo: glFloat;
    procedure DrawQuad;
    begin
      glBegin(GL_QUADS);
        glTexCoord2f(0 + d, 0 - d);
        glVertex2f(0, 0); //bottom left corner
        glTexCoord2f(0 + d, yo - d);
        glVertex2f(0, DisplayHeight());
        glTexCoord2f(xo + d, yo - d);
        glVertex2f(DisplayWidth(), DisplayHeight());
        glTexCoord2f(xo + d, 0 - d);
        glVertex2f(DisplayWidth(), 0);
      glEnd;
    end;
  begin
    glViewport(0, 0, DisplayWidth, DisplayHeight);
     glMatrixMode(GL_PROJECTION);
     glLoadIdentity();
     glMatrixMode(GL_MODELVIEW);
     glLoadIdentity();
    glOrtho(0, DisplayWidth(), 0, DisplayHeight(), 0 ,1);
    glDisable(GL_DEPTH_TEST);
    glDepthMask(GL_FALSE);
    glDisable(GL_BLEND);
    glDisable(GL_ALPHA_TEST);
    
    glColor4f(brightness1, brightness1, brightness1, 1);
    glBindTexture(GL_TEXTURE_2D, tex1);
    d:=blur1 * 1/tex1size;
    xo:=image1width/tex1size;
    yo:=image1height/tex1size;
    DrawQuad();

    glEnable(GL_BLEND);
    glBlendFunc(GL_ONE, GL_ONE);
    glColor4f(brightness2, brightness2, brightness2, brightness2);
    glBindTexture(GL_TEXTURE_2D, tex2);
    d:=blur2 * 1/tex2size;
    xo:=image2width/tex2size;
    yo:=image2height/tex2size;
    DrawQuad();
  end;
  
// --------------------------------------------------------------------------//


end.


{    for i:=0 to 3 do color1[i]:=brightness/2;
//    color1[3]:=1;

   // glDisable(GL_BLEND);
   // glBlendFunc(GL_ONE, GL_ONE);
glActiveTextureARB (GL_TEXTURE0_ARB);
glEnable (GL_TEXTURE_2D);
glBindTexture (GL_TEXTURE_2D,  tex1);
glTexEnvf(GL_TEXTURE_ENV, GL_COMBINE_RGB_ARB, GL_REPLACE);

glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE_ARB);
glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE0_RGB_ARB, GL_TEXTURE);
glTexEnvf(GL_TEXTURE_ENV, GL_OPERAND0_RGB_ARB, GL_PRIMARY_COLOR);
glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_RGB_ARB, GL_REPLACE);

glActiveTextureARB (GL_TEXTURE1_ARB);
glBindTexture (GL_TEXTURE_2D,  tex1);
glEnable (GL_TEXTURE_2D);
glTexEnvf(GL_TEXTURE_ENV, GL_COMBINE_RGB_ARB, GL_ADD);

glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE_ARB);
glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_RGB_ARB, GL_ADD);
glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE0_RGB_ARB, GL_TEXTURE);
glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE1_RGB_ARB, GL_PREVIOUS_ARB);
glTexEnvf(GL_TEXTURE_ENV, GL_OPERAND0_RGB_ARB, GL_PRIMARY_COLOR);
glTexEnvf(GL_TEXTURE_ENV, GL_OPERAND1_RGB_ARB, GL_ONE);

    glBegin(GL_QUADS);
      glMultiTexCoord2fARB(GL_TEXTURE0_ARB,0 + d, 0 + d);
      glMultiTexCoord2fARB(GL_TEXTURE1_ARB,0 - d, 0 - d);
      glVertex2f(0, 0); //bottom left corner

      glMultiTexCoord2fARB(GL_TEXTURE0_ARB,0 + d, yo + d);
      glMultiTexCoord2fARB(GL_TEXTURE1_ARB,0 - d, yo - d);
      glVertex2f(0, imageheight div 2);

      glMultiTexCoord2fARB(GL_TEXTURE0_ARB,xo + d, yo + d);
      glMultiTexCoord2fARB(GL_TEXTURE1_ARB,xo - d, yo - d);
      glVertex2f(imagewidth div 2, imageheight div 2);

      glMultiTexCoord2fARB(GL_TEXTURE0_ARB,xo + d, 0 + d);
      glMultiTexCoord2fARB(GL_TEXTURE1_ARB,xo - d, 0 - d);
      glVertex2f(imagewidth div 2, 0);
    glEnd;
    glActiveTextureARB(GL_TEXTURE1_ARB);
    glDisable(GL_TEXTURE_2D);
    glActiveTextureARB(GL_TEXTURE0_ARB);
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);}
    
    
    
    
    
    
    
    
    

{    glActiveTextureARB(GL_TEXTURE0_ARB);
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, tex1);
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
    glTexEnvfv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, @color1[0]);

    glActiveTextureARB(GL_TEXTURE1_ARB);
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, tex2);
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_ADD);
    glTexEnvfv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, @color2[0]);

    d1:=blur1 * 1/tex1size;
    d2:=blur2 * 1/tex2size;

    xo1:=image1width/tex1size;
    yo1:=image1height/tex1size;

    xo2:=image2width/tex2size;
    yo2:=image2height/tex2size;

    glBegin(GL_QUADS);
      glMultiTexCoord2fARB(GL_TEXTURE0_ARB,0 + d1, 0 + d1);
      glMultiTexCoord2fARB(GL_TEXTURE1_ARB,0 + d2, 0 + d2);
      glVertex2f(0, 0); //bottom left corner

      glMultiTexCoord2fARB(GL_TEXTURE0_ARB,0 + d1, yo1 + d1);
      glMultiTexCoord2fARB(GL_TEXTURE1_ARB,0 + d2, yo2 + d2);
      glVertex2f(0, DisplayHeight());

      glMultiTexCoord2fARB(GL_TEXTURE0_ARB,xo1 + d1, yo1 + d1);
      glMultiTexCoord2fARB(GL_TEXTURE1_ARB,xo2 + d2, yo2 + d2);
      glVertex2f(DisplayWidth(), DisplayHeight());

      glMultiTexCoord2fARB(GL_TEXTURE0_ARB,xo1 + d1, 0 + d1);
      glMultiTexCoord2fARB(GL_TEXTURE1_ARB,xo2 + d2, 0 + d2);
      glVertex2f(DisplayWidth(), 0);
    glEnd;

    glActiveTextureARB(GL_TEXTURE1_ARB);
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
    glDisable(GL_TEXTURE_2D);
    glActiveTextureARB(GL_TEXTURE0_ARB);
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE_EXT); }

